import importlib
import inspect
import logging
import sys
from pathlib import Path
from unittest.mock import Mock

import pytest

from aio_wx_widgets import widgets

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)


def load_module(module_file: Path):
    """Load and import a module based on the location of the *.py file."""
    try:
        name = module_file.stem
        spec = importlib.util.spec_from_file_location(name, module_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)
        return module
    except Exception as err:
        _LOGGER.exception(err)
        raise


def get_all_modules(package):
    """Yield all the modules from the provided package."""
    base = Path(inspect.getabsfile(package)).parent

    for fl in base.glob("*.py"):
        print(f"loading module {fl}")
        yield load_module(fl)


def get_all_classes_defined_in_module(module):
    """Return a tuple with the name of the class and the class object to be
    instantiated."""
    for _cls in inspect.getmembers(module, inspect.isclass):
        if module.__name__ == _cls[1].__module__:
            yield _cls


def get_all_widget_classes(package):
    for module in get_all_modules(package):
        for _cls in get_all_classes_defined_in_module(module):
            name, cls = _cls
            kwargs = {}
            if name == "AioButton":
                kwargs["label"] = "a label"
                kwargs["callback"] = Mock()
            elif name == "Group":
                kwargs["label"] = "grouplabel"
            elif name == "Entry":
                kwargs["binding"] = Mock()
            elif name == "Select":
                kwargs["choices"] = [Mock()]
            elif name == "CheckBox":
                kwargs["label"] = "checkbox_label"
                kwargs["binding"] = Mock()
            elif name == "LabelledItem":
                kwargs["label_text"] = "labeltext"
                kwargs["item"] = Mock()
            yield name, cls, kwargs


# # todo: make this a module based fixture.
# app = WxAsyncApp()


@pytest.mark.parametrize(
    "widget_class,kwargs",
    [
        pytest.param(_cls[1], _cls[2], id=_cls[0])
        for _cls in get_all_widget_classes(widgets)
    ],
)
def test_init(widget_class, kwargs, wx_app):
    widget_class(**kwargs)



"""Checkbox."""

import logging
from typing import Callable, Optional, Union

import wx

from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding, TwoWayBindable

_LOGGER = logging.getLogger(__name__)

__all__ = ["CheckBox"]


class CheckBox(BaseWidget):
    """Checkbox widget."""

    def __init__(
        self,
        label: str,
        binding: Optional[Binding],
        change_callback: Optional[Callable[[bool], None]] = None,
        enabled: Union[bool, Binding] = True,
        min_width=-1,
    ):
        """Init.

        Args:
            label: Checkbox label
            binding: property binding.
            change_callback: called when checkbox value changes.
        """

        value_binding = (
            TwoWayBindable(binding, self._get_ui_value, self._set_ui_value)
            if binding is not None
            else None
        )
        super().__init__(
            wx.CheckBox(),
            min_width=min_width,
            value_binding=value_binding,
            enabled=enabled,
        )
        self._label = label
        self._change_callback = change_callback

    def _set_ui_value(self, value):
        val = bool(value)
        _LOGGER.debug("Setting value to %s, %s", val, id(self))
        self.ui_item.SetValue(val)

        # setting this value manually as the SetValue command does -in this case-
        # not trigger the _on_ui_change callback.
        self._value_binding.fire_update_event = True

    def _get_ui_value(self, force: bool) -> bool:  # noqa
        val: bool = self.ui_item.GetValue()
        return val

    def _on_ui_change(self, *args, **kwargs):
        self._value_binding.on_ui_change(*args, **kwargs)

        if self._change_callback:
            self._change_callback(self._get_ui_value(True))

    def __call__(self, parent):
        self.ui_item.Create(parent, label=str(self._label))
        self.ui_item.Bind(wx.EVT_CHECKBOX, self._on_ui_change)

        self._init()

        return self



"""A labelled item."""

import logging

from aio_wx_widgets.containers.grid import Grid
from aio_wx_widgets.core.data_types import HorAlign, VerAlign
from aio_wx_widgets.widgets.text import Text

_LOGGER = logging.getLogger(__name__)

__all__ = ["LabelledItem"]


class LabelledItem(Grid):
    """Labelled item."""

    def __init__(
        self,
        label_text: str,
        item,
        align_right=True,
        label_weight=1,
        item_weight=1,
        item_alignment=None,
    ):
        super().__init__()
        self.label_text = label_text
        self._item = item
        self._label_weight = label_weight
        self._item_weight = item_weight
        self.align_right = align_right
        self._item_alignment = item_alignment

    def __call__(self, parent, *args, **kwargs):
        super().__call__(parent)
        if self.align_right:
            hor_alignment = HorAlign.right
        else:
            hor_alignment = None
        self.add(
            Text(self.label_text),
            weight=self._label_weight,
            hor_align=hor_alignment,
            ver_align=VerAlign.center,
        )
        self.add(self._item, weight=self._item_weight, hor_align=self._item_alignment)

        return self

    def __str__(self):
        return (
            f"{self.__class__.__name__}, "
            f"label_weigth({self._label_weight}), "
            f"item_weight({self._item_weight}), "
            f"align_right({self.align_right}), "
            f"item_alignment({self._item_alignment})"
        )



import logging

import pytest

from aio_wx_widgets.core.validators import (
    ValidationError,
    all_digits_validator,
    int_validator,
)

_LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize("incoming,expected", [("0", 0), ("1", 1), ("-1", -1)])
def test_int_validator(incoming, expected):

    result = int_validator(incoming, True)

    assert isinstance(result, int)
    assert result == expected


@pytest.mark.parametrize("incoming", ["a"])
def test_int_validator_fail(incoming):
    with pytest.raises(ValidationError):
        int_validator(incoming, True)


@pytest.mark.parametrize("incoming", ["0002340324"])
def test_all_digits_validator(incoming):
    result = all_digits_validator(incoming, True)

    assert result == incoming


@pytest.mark.parametrize("incoming", ["000234a0324"])
def test_all_digits_validator_fail(incoming):
    with pytest.raises(ValidationError):
        all_digits_validator(incoming, True)



"""Error messages.

Used in combination with validators.
"""

import logging

import wx

from aio_wx_widgets.const import ERROR_COLOR

_LOGGER = logging.getLogger(__name__)


class ErrorPopup(wx.PopupWindow):
    """A popup.

    Opens near the widget when a validation error occurs.
    """

    def __init__(self, parent, content, style=0):
        super().__init__(parent, style)
        panel = wx.Panel(self)
        panel.SetBackgroundColour(ERROR_COLOR)
        message = wx.StaticText(panel, -1, content)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(message, 0, wx.ALL, 5)
        panel.SetSizer(sizer)
        sizer.Fit(panel)
        sizer.Fit(self)
        self.Layout()



import logging
from unittest.mock import Mock

import pytest

from aio_wx_widgets.controller import BaseController
from aio_wx_widgets.core.binding import WATCHERS, Binding, TwoWayBindable

_LOGGER = logging.getLogger(__name__)
#
# class SomeBinding(Binding):
#


class SomeController(BaseController):
    def __init__(self, model):
        super().__init__(model)
        self.a_value = 10


@pytest.fixture
def some_controller(base_model):
    return SomeController(base_model)


@pytest.fixture
def some_binding(some_controller):
    return Binding(bound_object=some_controller, bound_property="a_value")


@pytest.fixture
def some_bindable(mocker, some_binding):
    get_ui_value = Mock()
    set_ui_value = Mock()
    # mocker.patch.object(TwoWayBindable, "_get_ui_value")
    # mocker.patch.object(TwoWayBindable, "_set_ui_value")

    return TwoWayBindable(some_binding, get_ui_value, set_ui_value)


def test_make_binding(some_bindable, some_controller):
    some_bindable.make_binding()

    # check if watcher is added to list of watchers.
    watchers = some_controller.__dict__[WATCHERS]["a_value"]
    assert len(watchers) == 1

    # Check if widget is updated with the default value.
    some_bindable._set_ui_value.assert_called_once_with(10)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from aio_wx_widgets.containers.grid import VERTICAL, Grid
from aio_wx_widgets.containers.group import Group, Section
from aio_wx_widgets.core.binding import Binding
from aio_wx_widgets.core.data_types import HorAlign, VerAlign
from aio_wx_widgets.core.validators import float_validator, int_validator
from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.checkbox import CheckBox
from aio_wx_widgets.widgets.labelled_item import LabelledItem
from aio_wx_widgets.widgets.text import Text
from aio_wx_widgets.widgets.text_entry import Entry

if TYPE_CHECKING:
    from demo.controller.controller_one import ControllerOne

_LOGGER = logging.getLogger(__name__)


class ViewOne(SimplePanel):
    def __init__(self, parent, controller: "ControllerOne"):
        self._controller = controller
        super().__init__(parent, scrollable=False)

    @property
    def controller(self) -> "ControllerOne":
        return self._controller

    def populate(self):
        """Populate this view."""

        # Use a context manager for container types like a group or grid.
        # A group is a container with a label and a sizer inside. Inside
        # this sizer widgets, or other containers can be placed.
        self.add(Text(binding=self.bind("int_val"), wrap=True), margin=20)
        with self.add(Grid()) as grd:
            grd.add(Text("This is long text", min_width=50))
            grd.add(Text("short", min_width=50))
        self.add(
            LabelledItem(
                "Label text",
                Entry(binding=self.bind("float_val"), validator=float_validator),
            )
        )

        with self.add(Section("Float validators")) as sc:
            sc.add(
                LabelledItem(
                    "Enter a float value",
                    Entry(binding=self.bind("float_val"), validator=float_validator,),
                    item_weight=2,
                )
            )
            with sc.add(Grid()) as grd:
                # the binding binds to an attribute defined in the controller
                # the weight determines how much space a specific item should consume
                # with respect to the other members of the container.

                grd.add(
                    Entry(binding=self.bind("float_val"), validator=float_validator,),
                    weight=4,
                )
                grd.add(
                    Text(binding=self.bind("float_val")),
                    weight=4,
                    ver_align=VerAlign.center,
                )
        with self.add(Section("Int validators")) as sc:
            sc.add(Text("Only integer values are allowed."))
            with sc.add(Grid()) as grd:
                grd.add(
                    Entry(binding=self.bind("int_val"), validator=int_validator,),
                    weight=3,
                    margin=0,
                )
                grd.add(
                    Entry(binding=self.bind("int_val"), validator=int_validator,),
                    weight=1,
                    margin=0,
                )
                grd.add(Text(binding=self.bind("int_val")), weight=1, margin=0)

        with self.add(Grid()) as grd:
            grd.add(Entry(binding=Binding(self._controller, "a_string_value")))
            grd.add(Entry(binding=Binding(self._controller, "a_string_value")))

        with self.add(Grid()) as grd:
            grd.add(AioButton("button one", self._set_value), weight=6)
            grd.add(AioButton("button two,self", self._set_value), weight=6)

        with self.add(Group("Nesting of grids")) as grp:
            with grp.add(Grid()) as grd:
                grd.add(
                    Text(text="Right aligned text"),
                    weight=3,
                    hor_align=HorAlign.right,
                    ver_align=VerAlign.center,
                )
                with grd.add(Grid(VERTICAL), weight=6, margin=0) as vert_grid:
                    vert_grid.add(
                        AioButton("Left aligned button", self._set_value),
                        hor_align=HorAlign.left,
                        margin=4,
                    )
                    vert_grid.add(AioButton("Set number entries.", self._set_value))
                    vert_grid.add(
                        Text(text="Center aligned text with a large margin."),
                        margin=(10, 10, 30, 5),
                        hor_align=HorAlign.center,
                    )
                    vert_grid.add(
                        AioButton(
                            "right aligned button.",
                            self._set_value,
                            enabled=self.bind("a_checkbox_value"),
                        ),
                        weight=2,
                        hor_align=HorAlign.right,
                    )
        self.add(CheckBox("A checkbox", binding=self.bind("a_checkbox_value")))
        self.add(
            CheckBox("The same checkbox value", binding=self.bind("a_checkbox_value"))
        )
        self.add(AioButton("open other window", self._on_open_second_window))
        self.add(AioButton("open third window", self._on_open_third_window))

    async def _on_open_second_window(self, evt):
        await self._controller.open_other_window()

    async def _on_open_third_window(self, evt):
        await self._controller.open_third_window()

    async def _on_open(self, evt):
        await self._controller.open_other_window()

    async def _set_value(self: ViewOne):
        await self._controller.set_value()



"""Base widget"""

from __future__ import annotations

import logging
from typing import Union

from aio_wx_widgets import type_annotations as T  # noqa
from aio_wx_widgets.core.binding import Binding, OneWayBindable, TwoWayBindable

_LOGGER = logging.getLogger(__name__)


class BaseWidget:
    """All widgets should inherit from this."""

    def __init__(
        self,
        ui_item: T.wx,
        min_width: int,
        value_binding: Union[OneWayBindable, TwoWayBindable, None],
        enabled: Union[bool, Binding] = True,
    ):
        self._enabled = enabled
        self.ui_item = ui_item
        self._value_binding = value_binding
        self._min_width = min_width

        self._enabled_binding = None
        if isinstance(self._enabled, Binding):
            self._enabled_binding = OneWayBindable(
                self._enabled, self._set_enabled_value
            )

    def _set_enabled_value(self, enabled):
        if enabled:
            self.ui_item.Enable()
        else:
            self.ui_item.Disable()

    def _make_bindings(self):
        if self._value_binding:
            self._value_binding.make_binding()
        if self._enabled_binding:
            self._enabled_binding.make_binding()

    def _init(self):
        self._make_bindings()
        if self._min_width > -1:
            actual_x_value = self.ui_item.GetSize()[0]
            if actual_x_value < self._min_width:
                self.ui_item.SetMinSize((self._min_width, -1))

    def __call__(self, parent):
        raise NotImplementedError



"""Datatypes used by widgets."""

import logging
from enum import Enum
from typing import Any, NamedTuple

import wx

_LOGGER = logging.getLogger(__name__)


class Choices(NamedTuple):
    """Choice item."""

    label: str
    value: Any


class HorAlign(Enum):
    """Horizontal alignment options."""

    left = wx.ALIGN_LEFT
    right = wx.ALIGN_RIGHT
    center = wx.ALIGN_CENTER_HORIZONTAL


class VerAlign(Enum):
    """Vertical alignment options."""

    top = wx.ALIGN_TOP
    bottom = wx.ALIGN_BOTTOM
    center = wx.ALIGN_CENTER_VERTICAL



"""WX button."""

import logging
from asyncio import iscoroutinefunction
from typing import Awaitable, Callable, Union

import wx
from wxasync import AsyncBind  # type: ignore

from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding

_LOGGER = logging.getLogger(__name__)

__all__ = ["AioButton"]


class AioButton(BaseWidget):
    """Button widget.

    Args:
        label: button label.
        callback: a coroutine function or normal function. Called when pressed.
    """

    def __init__(
        self,
        label: str,
        callback: Union[
            Callable[[object, wx.Event], Awaitable], Callable[[object, wx.Event], None]
        ],
        enabled: Union[bool, Binding] = True,
        min_width=-1,
    ):
        super().__init__(
            wx.Button(), min_width=min_width, value_binding=None, enabled=enabled
        )
        self._label = label
        self._call_back = callback

    @property
    def label(self):
        """Return the button label."""
        return self._label

    @label.setter
    def label(self, value: str):
        self._label = str(value)
        self.ui_item.SetLabelText(str(self._label))

    def __call__(self, parent):
        self.ui_item.Create(parent, label=str(self._label))
        if iscoroutinefunction(self._call_back):

            AsyncBind(wx.EVT_BUTTON, self._call_back, self.ui_item)
        else:
            self.ui_item.Bind(wx.EVT_BUTTON, self._call_back)

        # self._make_bindings()
        self._init()
        return self



import logging

from demo.controller import controller_two
from demo.model.demo_model import DemoModel
from demo.views.view_two import ViewTwo

from aio_wx_widgets.frame import DefaultFrame

_LOGGER = logging.getLogger(__name__)


class SecondWindow(DefaultFrame):
    def __init__(self):
        super().__init__("second window")

        model = DemoModel()
        controller = controller_two.ControllerTwo(model)
        view = ViewTwo(self, controller)
        view.populate()



import logging

from demo.controller.controller_one import ControllerOne
from demo.model.demo_model import DemoModel
from demo.views.view_one import ViewOne

from aio_wx_widgets.frame import DefaultFrame

_LOGGER = logging.getLogger(__name__)


class MainWindow(DefaultFrame):
    def __init__(self):
        super().__init__("Main window")

        model = DemoModel()
        controller = ControllerOne(model)
        view = ViewOne(self, controller)
        view.populate()



"""Text widgets."""

import logging
from typing import Optional, Union

import wx

from aio_wx_widgets.colors import GREEN
from aio_wx_widgets.const import is_debugging
from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding, OneWayBindable

_LOGGER = logging.getLogger(__name__)

__all__ = ["Text"]


def _get_font_info(current_font: wx.Font, font_size: float = 1, bold=False) -> wx.Font:
    font = wx.Font()

    if font_size != 1:
        current_font_size = current_font.GetPointSize()
        new_font_size = int(font_size * current_font_size)
        font.SetPointSize(new_font_size)

    if bold:
        font.MakeBold()

    return font


class Text(BaseWidget):
    """Static text."""

    def __init__(
        self,
        text="",
        font_size: float = 1,
        color: Optional[int] = None,
        enabled: Union[bool, Binding] = True,
        binding: Optional[Binding] = None,
        bold=False,
        wrap=False,
        min_width=-1,
    ):
        """Init.

        Args:
            text:
            font_size:
        """
        value_binding = (
            OneWayBindable(binding, self._set_ui_value) if binding is not None else None
        )

        super().__init__(
            wx.StaticText(),
            min_width=min_width,
            value_binding=value_binding,
            enabled=enabled,
        )
        self._parent = None
        self._color = color
        self._text = text
        self._wrap = wrap

        self._font = _get_font_info(self.ui_item.GetFont(), font_size, bold=bold)

        self._previous_size = None

    def _set_ui_value(self, value):
        self.set_text(value)

    def __call__(self, parent):
        self._parent = parent
        self.ui_item.Create(parent)

        if self._font:
            self.ui_item.SetFont(self._font)

        if self._text:
            self.set_text(self._text, self._color)
        if self._value_binding:
            self.set_text(self._value_binding.get_property_value(), self._color)
        if is_debugging():
            self.ui_item.SetBackgroundColour(GREEN)
        self._init()
        self.ui_item.Bind(wx.EVT_SIZE, self._on_size)

        return self

    def _set_text(self, color=None, client_size=None):
        # _LOGGER.debug("Setting text to: %s", self._text)
        self.ui_item.Unbind(wx.EVT_SIZE)

        self.ui_item.SetLabel(str(self._text))

        if self._wrap:
            if client_size:
                self.ui_item.Wrap(client_size[0])

        if color:
            self.ui_item.SetForegroundColour(color)

        self.ui_item.Bind(wx.EVT_SIZE, self._on_size)

    def _on_size(self, evt):
        size = evt.Size
        if self._previous_size is not None and self._previous_size[0] == size[0]:
            return

        self._previous_size = size
        self._set_text(client_size=size)

        # evt.Skip()

    def set_text(self, text, color=None):
        """Set or change the text."""
        self._text = str(text)
        self._set_text(color)



"""Group layout item. Draws a line around children and sets a title."""


import logging

import wx

from aio_wx_widgets.core.sizers import SizerMixin
from aio_wx_widgets.widgets import text

_LOGGER = logging.getLogger(__name__)

__all__ = ["Group", "Section"]


class Group(SizerMixin):
    """A Group widget.

    +--<LABEL>-------------+
    |                      |
    |  - Widget 1          |
    |  - Widget 2          |
    |                      |
    +----------------------+

    """

    default_sizer_margin = 20

    def __init__(self, label, **kwargs):
        """Init.

        Args:
            label:
        """
        super().__init__()
        self._label = label
        self._extra_bottom_border = 10
        self._sizer_ = wx.BoxSizer(wx.VERTICAL)
        self._ui_item = wx.StaticBox()
        kwargs["sizer"] = self._sizer
        self._bottom_border = 1

    @property
    def _sizer(self):
        return self._sizer_

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    def __call__(self, parent):
        """Call this class as a function.

        Args:
            parent: The parent of this ui item.

        """
        self.ui_item.Create(parent, label=self._label)

        top, other = self.ui_item.GetBordersForSizer()
        self._bottom_border = other + self._extra_bottom_border
        self._sizer.AddSpacer(top + 5)
        self.ui_item.SetSizer(self._sizer)

        return self

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, *args):
        """Exit context manager."""
        self._finish()

    def _finish(self):
        self._sizer.AddSpacer(self._bottom_border)


class Section(SizerMixin):
    """A section with a large header text and an optional closing line.

    LARGE HEADER TEXT

    - content
    - content
    ----------------------------------- [optional closing line]
    """

    def __init__(self, header: str, add_closing_line=True):
        """Init."""
        super().__init__()
        self._ui_item = self._sizer_ = wx.BoxSizer(orient=wx.VERTICAL)
        self._parent = None
        self._header = header
        self._add_closing_line = add_closing_line

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    @property
    def _sizer(self):
        return self._sizer_

    def __call__(self, parent):
        self._parent = parent
        return self

    def __enter__(self):
        """Enter the context manager."""
        self.add(
            text.Text(text=self._header, font_size=1.5, bold=True), margin=(0, 0, 15, 0)
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager"""
        self.add(wx.StaticLine(self._parent), create=False, margin=(0, 0, 15, 1))



"""Text entry widgets."""

# pylint: disable=invalid-name,no-self-use
from __future__ import annotations

import logging
from typing import Any, Callable, Optional, Union

import wx

from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding, TwoWayBindable
from aio_wx_widgets.core.error_message import ErrorPopup
from aio_wx_widgets.core.validators import ValidationError

_LOGGER = logging.getLogger(__name__)

__all__ = ["Entry"]


class Entry(BaseWidget):
    """A general textctrl for any input."""

    def __init__(
        self,
        binding: Binding,
        label: Union[None, int, str] = None,
        validator: Callable[[Any, bool], Any] = None,
        on_change: Callable[[Entry, Any], None] = None,
        min_width=-1,
        enabled: Union[bool, Binding] = True,
    ):
        super().__init__(
            wx.TextCtrl(),
            min_width=min_width,
            value_binding=TwoWayBindable(
                binding,
                self._get_ui_value,
                self._set_ui_value,
                display_error=self.display_error,
            ),
            enabled=enabled,
        )
        self.value: Optional[int] = None
        self._label = label
        self._txt = self.ui_item
        self._validator = validator
        self._popup: Optional[ErrorPopup] = None
        self._allow_none = True

        self.ui_item.Bind(wx.EVT_KILL_FOCUS, self._on_focus_lost)
        self._on_change = on_change

    def _on_focus_lost(self, evt):
        _LOGGER.debug("Lost focus on element.")
        self._value_binding.update_binding(force=True)
        evt.Skip()

    def _set_ui_value(self, value):
        if value is None:
            value = ""
        _LOGGER.debug(
            "Updating text entry with value: %s, instance %s", value, id(self)
        )
        self._txt.SetValue(str(value))
        _LOGGER.debug("Done updating.")

    def _kill_popup(self):
        if self._popup:
            _LOGGER.debug("Destroying popup")
            self._popup.Destroy()
            self._popup = None

    def _get_ui_value(self, force: bool) -> Any:
        self._kill_popup()
        val = self._txt.GetValue()

        if val == "" and self._allow_none:
            assert self._value_binding is not None
            _LOGGER.debug(
                "Setting property to None, %s",
                self._value_binding.binding.bound_property,
            )
            return None

        if self._validator:
            return self._validator(self._txt.GetValue(), force)
        return self._txt.GetValue()

    def _on_ui_change(self, *args, **kwargs):
        self._value_binding.on_ui_change(*args, **kwargs)

        if self._on_change:
            self._on_change(self, self._get_ui_value(True))

    def __call__(self, parent):
        args = dict({"parent": parent})
        if self._label:
            args["value"] = str(self._label)

        self._txt.Create(**args)
        # self.ui_item.SetSizeHints(self._min_width, -1)
        self._txt.Bind(wx.EVT_TEXT, self._on_ui_change)

        # self._make_bindings()
        super()._init()
        return self

    def display_error(self, exception: ValidationError):
        """Display a validation error."""
        self._kill_popup()

        message = str(exception)
        self._popup = ErrorPopup(self.ui_item, message, style=wx.SIMPLE_BORDER)

        pos = self.ui_item.ClientToScreen((0, 0))
        sz = self.ui_item.GetSize()
        self._popup.Position(pos, (0, sz[1]))
        self._popup.Show()
        self.ui_item.SetFocus()



from __future__ import annotations

import logging
from random import randint
from typing import Sequence

from events import Events

from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.controller import BaseController
from aio_wx_widgets.core.data_types import Choices

_LOGGER = logging.getLogger(__name__)


class ControllerTwo(BaseController):
    def __init__(self, model):
        self.value_1: int = 0
        self.a_string_value = "A certain string"
        super().__init__(model)
        # self.create_task(self.value_setter())
        # self.add_to_log = Channel("Log messages")
        self.add_to_log = Events()
        self.choices: Sequence[T.Choice] = [
            Choices(label="choice 1", value={"value": 1}),
            Choices(label="choice 2", value={"value": 2}),
        ]
        self.selected_item = self.choices[0]

    async def set_value(self):
        val = randint(1, 100)
        _LOGGER.debug("Setting bound property to %s", val)
        self.value_1 = val

    # async def value_setter(self):
    #     try:
    #         while 1:
    #             val = str(randint(1, 100))
    #             _LOGGER.debug("Setting value to : %s", val)
    #             self.add_to_log.publish(val)
    #             await asyncio.sleep(2)
    #     except CancelledError:
    #         _LOGGER.debug("Stopping value setter task.")



"""Sizer and Layout tooling."""
from __future__ import annotations

import collections
import logging
from typing import List, Optional, Tuple, Union

import wx

from aio_wx_widgets import type_annotations as T  # noqa
from aio_wx_widgets.core.binding import Binding
from aio_wx_widgets.core.data_types import HorAlign, VerAlign

_LOGGER = logging.getLogger(__name__)

# T_var = TypeVar("T_var")


def _make_window(item: wx.Window, margins: List[Tuple[int, int]]):
    val = margins[0][1]
    for margin in margins[1:]:
        val = val | margin[1]

    new_sizer = wx.BoxSizer()
    margin_value = margins[0][0]
    new_sizer.Add(item, 1, val, margin_value)
    return new_sizer


def _margin_wrapper(item, margins: List[Tuple[int, int]]) -> wx.Window:
    if len(margins) == 0:
        return item

    first = margins.pop(0)

    values = [first]
    for idx, _ in enumerate(margins):
        if margins[idx][0] == first[0]:
            values.append(margins.pop(idx))

    item = _make_window(item, values)
    return _margin_wrapper(item, margins)


def _align_item(
    item: Union[wx.Window, wx.BoxSizer],
    current_sizer_orientation,
    hor_align: Optional["HorAlign"],
    ver_align: Optional[VerAlign],
    current_layout: int,
) -> Tuple[Union[wx.Window, wx.BoxSizer], int]:

    if current_sizer_orientation == wx.VERTICAL:
        _LOGGER.debug("Current sizer is vertical")
        if hor_align:
            _LOGGER.debug("Item has hor alignment set.")
            current_layout = current_layout | hor_align.value
            return _align_item(
                item, current_sizer_orientation, None, ver_align, current_layout
            )
    if current_sizer_orientation == wx.HORIZONTAL:
        _LOGGER.debug("Current sizer is horizontal.")
        if ver_align:
            _LOGGER.debug("Item has ver alignment set.")
            current_layout = current_layout | ver_align.value
            return _align_item(
                item, current_sizer_orientation, hor_align, None, current_layout
            )
    if hor_align:
        _LOGGER.debug(
            "Setting a horizontal alignment but current orientation is %s",
            current_sizer_orientation,
        )
        sizer = wx.BoxSizer(orient=wx.VERTICAL)
        sizer.Add(item, 0, hor_align.value, 0)
        return _align_item(sizer, wx.VERTICAL, None, ver_align, current_layout)

    if ver_align:
        sizer = wx.BoxSizer()
        sizer.Add(item, 0, ver_align.value, 0)
        return _align_item(sizer, wx.HORIZONTAL, hor_align, None, current_layout)
    return item, current_layout


def _add(
    item: T.Widget,
    parent,
    sizer,
    weight,
    margin,
    default_margin,
    create,
    hor_align: Optional[HorAlign],
    ver_align: Optional[VerAlign],
) -> T.Widget:
    if create:
        # this is an item which is part of the aio_wx_widgets family.
        # It is assumed it has the ui_item property.
        item = item(parent)
        ui_item = item.ui_item
    else:
        # this is assumed to be a "normal" wx-widgets item.
        ui_item = item
    if margin is None:
        margin = default_margin
    elif isinstance(margin, collections.Sequence):
        margins = [
            (margin[0], wx.LEFT),
            (margin[1], wx.RIGHT),
            (margin[2], wx.TOP),
            (margin[3], wx.BOTTOM),
        ]
        ui_item = _margin_wrapper(ui_item, margins)
        margin = 0

    layout = wx.ALL

    _LOGGER.debug(
        ">> Aligning %s, ver_align(%s), hor_align(%s)", type(item), ver_align, hor_align
    )
    if hor_align is None and ver_align is None:

        layout = layout | wx.EXPAND
    else:
        ui_item, layout = _align_item(
            ui_item, sizer.Orientation, hor_align, ver_align, layout
        )

    sizer.Add(ui_item, weight, layout, margin)

    return item


class SizerMixin:
    """Mixin providing sizer functionality"""

    default_sizer_margin = 5

    def __init__(self):
        self._parent = None
        super().__init__()

    def add(
        self,
        item: T.Widget,
        weight=0,
        margin=(5, 5, 1, 1),  # left,right,top,bottom
        create=True,
        hor_align: Optional[HorAlign] = None,
        ver_align: Optional[VerAlign] = None,
    ) -> T.Widget:
        """Add an item to this panel

        Args:
            item: the item to be added.
            weight: The relative weight to other items in this container and the size
                it should take
            margin: An int to use one margin for all sides.
                A tuple (left,right,top,bottom) for side specific margins.
            create: A ready made component can be added to.
            hor_align: Horizontal alignment.
            ver_align: Vertical alignment.
        """
        try:
            parent = self._parent  # type: ignore
            if parent is None:
                raise AttributeError("parent cannot be none.")
        except AttributeError:
            parent = self.ui_item  # type: ignore

        return _add(
            item,
            parent,
            self._sizer,  # type: ignore
            weight,
            margin,
            SizerMixin.default_sizer_margin,
            create,
            hor_align=hor_align,
            ver_align=ver_align,
        )

    @property
    def ui_item(self):
        """Return UI item."""
        raise NotImplementedError()

    @property
    def _sizer(self):
        raise NotImplementedError()

    def add_space(self, proportion=1):
        """Add a stretching spacer."""
        self._sizer.AddStretchSpacer(prop=proportion)


class PanelMixin:
    """Mixin providing extra functionality."""

    def __init__(self):
        self.ui_item.Bind(wx.EVT_WINDOW_DESTROY, self._on_close)
        super().__init__()

    def _on_close(self, evt):
        _LOGGER.debug("Window destroyed.")
        self.controller.deactivate()
        evt.Skip()

    @property
    def controller(self) -> T.BaseController:
        """Return the controller for this panel."""
        raise NotImplementedError()

    @property
    def ui_item(self):
        """Return UI item."""
        raise NotImplementedError()

    def bind(self, prop: str):
        """Bind a property to a bindable control.

        Args:
            prop: the attribute/property to bind.
        """
        return Binding(self.controller, prop)

    def populate(self):
        """Add and configure widgets here."""



from __future__ import annotations

import logging

from aio_wx_widgets.containers.grid import Grid
from aio_wx_widgets.containers.group import Group
from aio_wx_widgets.core.binding import Binding
from aio_wx_widgets.core.sizers import PanelMixin, SizerMixin
from aio_wx_widgets.panels.splitter_panel import TwoSplitterWindow
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.text import Text
from aio_wx_widgets.widgets.text_entry import Entry

_LOGGER = logging.getLogger(__name__)


class SplitterWindow(TwoSplitterWindow, PanelMixin, SizerMixin):
    def __init__(self, parent, controller, scrollable=False, window_one_width=250):
        super().__init__(
            parent, splitter_one_scrollable=False, splitter_two_scrollable=False
        )
        self._controller = controller
        # self.splitter_window_two.add(Text("This is text"))
        self.splitter_window_two.add(Entry(self.bind("value_1")))

    @property
    def controller(self):
        return self._controller


class ViewThree(SplitterWindow):
    def __init__(self, parent, controller):
        self._controller = controller
        super().__init__(parent, controller, scrollable=False)

        self.split_button = AioButton("Remove split", self._toggle_window_2)
        self._toggle_state = True

    @property
    def controller(self):
        return self._controller

    def populate(self):
        """Populate this view."""
        self.splitter_window_one.add(Entry(binding=self.bind("value_1")), margin=20)
        # Use a context manager for container types like a group or grid.
        # A group is a container with a label and a sizer inside. Inside
        # this sizer widgets, or other containers can be placed.
        with self.splitter_window_one.add(Group("A labelled container.")) as group:
            # group.add(Text(text="A horizontal grid."))

            with group.add(Grid()) as grd:
                # the binding binds to an attribute defined in the controller
                # the weight determines how much space a specific item should consume
                # with respect to the other members of the container.
                grd.add(Entry(binding=self.bind("value_1")), weight=3, margin=3)
                grd.add(Entry(binding=self.bind("value_1")), weight=1, margin=3)
                grd.add(Entry(binding=self.bind("value_1")), weight=1, margin=3)

        with self.add(Group("Any text entry.")) as group:
            group.add(Entry(binding=Binding(self._controller, "a_string_value")))
            group.add(Entry(binding=Binding(self._controller, "a_string_value")))

        with self.add(Grid()) as grd:
            grd.add(AioButton("button one", self._set_value), weight=6)
            grd.add(AioButton("button two,self", self._set_value), weight=6)
        with self.add(Grid(), margin=20) as grd:
            grd.add(
                Text(
                    "This is a long text that actually should wrap properly and right from the start. Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum. This is the last text.",
                    wrap=True,
                ),
                weight=1,
                margin=0,
            )

        self._btn = AioButton("toggle", self._toggle)
        with self.add(Grid()) as grd:
            grd.add(AioButton("toggle", self._toggle))
            grd.add(self._btn)

        # self.add_space(2)

        # self.add(
        #     self.split_button, hor_align=HorAlign.right,
        # )
        self._set_split_button_text()

    def _toggle(self, evt):
        self._toggle_state = not self._toggle_state
        if self._toggle_state:
            self._btn.disable()
        else:
            self._btn.enable()

    def _set_split_button_text(self):
        if self.is_split:
            self.split_button.label = "Close"
        else:
            self.split_button.label = "Open"

    def _toggle_window_2(self, evt):
        if self.is_split:
            self.hide_window2()
        else:
            self.show_window2()
        self._set_split_button_text()

    async def _on_open(self, evt):
        await self._controller.open_other_window()

    async def _set_value(self, evt):
        await self._controller.set_value()



"""Panel. Sits in a frame. Holds widgets."""
from __future__ import annotations

import logging

import wx
from wx.lib.scrolledpanel import ScrolledPanel

from aio_wx_widgets import type_annotations as T  # noqa
from aio_wx_widgets.core.sizers import PanelMixin, SizerMixin

_LOGGER = logging.getLogger(__name__)


class SimplePanel(PanelMixin, SizerMixin):
    """A simple panel."""

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    @property
    def _sizer(self):
        return self._sizer_

    def __init__(self, parent, scrollable=False, **kwargs):
        """Init."""

        if scrollable:
            self._ui_item = ScrolledPanel(parent)
            self._ui_item.SetupScrolling()
        else:
            self._ui_item = wx.Panel(parent)
        super().__init__()

        self._sizer_ = wx.BoxSizer(wx.VERTICAL)
        self.ui_item.SetSizer(self._sizer)

    @property
    def controller(self) -> T.BaseController:
        """Return the controller."""
        raise NotImplementedError()



"""All type annotations go here."""


from typing import TYPE_CHECKING, Any, TypeVar, Union

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop

    import wx

    from aio_wx_widgets.controller import BaseController
    from aio_wx_widgets.core.base_widget import BaseWidget
    from aio_wx_widgets.core.binding import Binding
    from aio_wx_widgets.core.sizers import SizerMixin

    try:
        import Protocol  # type: ignore
    except ImportError:
        from typing_extensions import Protocol

    assert BaseController
    assert AbstractEventLoop
    assert Binding
    assert wx
    assert BaseWidget

    class Choice(Protocol):
        """Typing protocol to be used for static type checking."""

        label: str
        value: Any

    T = TypeVar("T")  # pylint: disable=invalid-name

    Widget = Union[BaseWidget, wx.Window, wx.BoxSizer, SizerMixin]



"""Constants."""

import logging
import os

_LOGGER = logging.getLogger(__name__)

ERROR_COLOR = "#FF8C8C"


def is_debugging() -> bool:
    """Check if app is in debugging mode."""
    if os.getenv("DEBUGGING") == "1":
        return True
    return False



"""Widgets."""

from aio_wx_widgets.widgets.button import *  # noqa: F403,F401
from aio_wx_widgets.widgets.checkbox import *  # noqa: F403,F401
from aio_wx_widgets.widgets.labelled_item import *  # noqa: F403,F401
from aio_wx_widgets.widgets.select import *  # noqa: F403,F401
from aio_wx_widgets.widgets.text import *  # noqa: F403,F401
from aio_wx_widgets.widgets.text_entry import *  # noqa: F403,F401



"""MVC controller."""
from __future__ import annotations

import asyncio
import logging
from asyncio import Task
from typing import Awaitable, List

# pylint: disable=unused-import
from aio_wx_widgets import type_annotations as T  # noqa
from aio_wx_widgets.core.binding import WATCHERS

_LOGGER = logging.getLogger(__name__)


class BaseController:
    """Base implementation of a controller."""

    def __init__(self, model: T.T):
        """Init.

        Args:
            model: The model.
        """

        self._model: T.T = model
        self._tasks: List[Task] = []

    @property
    def model(self):
        """Return the model of this controller."""
        return self._model

    def __setattr__(self, item, value):
        # print(f"Setting item {item} to value {value}")
        # self.__dict__[item] = value
        # setattr(self,item,value)
        super().__setattr__(item, value)
        if WATCHERS not in self.__dict__:
            return
        if item not in self.__dict__[WATCHERS]:
            return

        watchers = self.__dict__[WATCHERS][item]
        for watcher in watchers:
            watcher(value)

    @property
    def _loop(self) -> "T.AbstractEventLoop":
        return asyncio.get_running_loop()

    def create_task(self, coro: Awaitable, callback=None) -> Task:
        """Create a task.

        When this step is deactivated, the task
        will be cancelled.
        """

        task: Task = self._loop.create_task(coro)
        if callback:
            task.add_done_callback(callback)
        self._tasks.append(task)
        return task

    def add_task(self, task):
        """Attach a task to this step.

        When step is deactivated, the
        task will be cancelled.
        """
        self._tasks.append(task)
        return task

    def deactivate(self):
        """Run on closing this step."""

        for task in self._tasks:
            task.cancel()



"""Property binding."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional

from aio_wx_widgets.core.validators import ValidationError

if TYPE_CHECKING:
    from aio_wx_widgets.controller import BaseController

_LOGGER = logging.getLogger(__name__)

WATCHERS = "_watchers_"


class Binding(NamedTuple):
    """Binding data."""

    bound_object: "BaseController"
    bound_property: "str"


class OneWayBindable:
    """A one way (read only) property binding.

    A bound UI will only receive values from a property. The ui itself is not able
    to update the property.
    """

    def __init__(self, binding: Optional[Binding], set_ui_value: Callable[[Any], None]):
        """Read only binding."""
        self._binding = binding
        self._set_ui_value = set_ui_value
        self._allow_none = False
        # Flag to stop firing updates when a property value has changed.
        # Prevents infinite loop of updating the value and firing update events.
        self.fire_update_event = True

    @property
    def binding(self):
        """Return the binding values."""
        return self._binding

    def make_binding(self):
        """Create the binding."""
        if self._binding is None:
            _LOGGER.debug("No binding defined.")
            return

        if WATCHERS not in self._binding.bound_object.__dict__:
            self._binding.bound_object.__dict__[WATCHERS] = {}
        if (
            self._binding.bound_property
            not in self._binding.bound_object.__dict__[WATCHERS]
        ):
            _LOGGER.debug("No watchers yet for %s", self._binding.bound_property)
            self._binding.bound_object.__dict__[WATCHERS][
                self._binding.bound_property
            ] = []
        self._binding.bound_object.__dict__[WATCHERS][
            self._binding.bound_property
        ].append(self._update)

        value = self.get_property_value()
        if value is None:
            _LOGGER.debug(
                "Original property has value None %s", self._binding.bound_property
            )
            self._allow_none = True

        self.fire_update_event = False
        self._set_ui_value(value)

    def _update(self, value):
        """Called when a property changes from outside this control.

        Update the control with the new value
        """
        _LOGGER.debug("Updating entry with new value.")
        self._set_ui_value(value)
        _LOGGER.debug("Updated.")

    def get_property_value(self):
        """Return the property value of the bound object.

        This would be the attribute defined in the controller.
        """
        return getattr(self._binding.bound_object, self._binding.bound_property)


class TwoWayBindable(OneWayBindable):
    """Base class for binding a property to a widget.

    Make a widget inherit from this. The binding property modifies the controller bound
    property adding a watcher and emitting change events.
    """

    def __init__(
        self, binding: Binding, get_ui_value, set_ui_value, display_error=None,
    ):

        super().__init__(binding, set_ui_value)

        self._get_ui_value = get_ui_value
        self._display_error = display_error

    def _set_property_value(self, value):
        setattr(self._binding.bound_object, self._binding.bound_property, value)

    def _update(self, value):
        """Called when a property changes from outside this control.

        Update the control with the new value
        """
        self.fire_update_event = False
        _LOGGER.debug("Setting fire event to False on %s", id(self))
        super()._update(value)

    def on_ui_change(self, *args, **kwargs):  # noqa
        """Callback when UI widget value changes by user input."""
        _LOGGER.debug("UI value changed on id %s", id(self))
        if not self.fire_update_event:
            _LOGGER.debug("Not updating bindings as fire_update_event is False")
            _LOGGER.debug("Setting fire event to True on %s", id(self))
            self.fire_update_event = True
            return
        _LOGGER.debug("Trigger the binding")

        self.update_binding()

        _LOGGER.debug("Firing update")

    def update_binding(self, force=False):
        """Update the binding property.

        Args:
            force: [bool] this will force the value to be cast into the value according
                to the validator.
        """
        try:
            val = self._get_ui_value(force)
        except ValidationError as err:
            if self._display_error:
                self._display_error(err)
        else:
            self._set_property_value(val)
            _LOGGER.debug("Setting fire event to True on %s", id(self))
            self.fire_update_event = True



import logging

import pytest

from aio_wx_widgets.core.data_types import HorAlign
from aio_wx_widgets.core.sizers import _align_item, wx

_LOGGER = logging.getLogger(__name__)


class MockBoxSizer:
    def __init__(self, orient=wx.HORIZONTAL):
        self.orient = orient
        self.item = None
        self.weight = None
        self.layout = None

    def Add(self, item, weight, orientation, layout):
        self.item = item
        self.weight = weight
        self.orient = orientation
        self.layout = layout


@pytest.fixture
def box_sizer(mocker):
    mocker.patch.object(wx, "BoxSizer", MockBoxSizer)


class UiItem:
    """A mocked ui item"""


def test_align_no_alignment(box_sizer):
    al = _align_item(UiItem(), wx.VERTICAL, None, None, current_layout=0)

    assert isinstance(al[0], UiItem)


def test_align_hor_alignment_vert_sizer(box_sizer):
    """No sizer used for wrapping the ui"""
    al = _align_item(UiItem(), wx.VERTICAL, HorAlign.center, None, current_layout=0)
    assert isinstance(al[0], UiItem)


def test_aling_hor_alignment_hor_sizer(box_sizer):
    """A sizer will be used to make the horizontal alignment happen."""

    al = _align_item(UiItem(), wx.HORIZONTAL, HorAlign.right, None, current_layout=0)

    assert isinstance(al[0], MockBoxSizer)
    assert isinstance(al[0].item, UiItem)



import logging

import wx

from aio_wx_widgets.core.sizers import _margin_wrapper
from aio_wx_widgets.frame import DefaultFrame
from aio_wx_widgets.panels.panel import SimplePanel

_LOGGER = logging.getLogger(__name__)


class MockWxWindow:
    def __init__(self, parent, *args, **kwargs):
        """Dummy init"""
        self.parent = parent
        self.sizer = None

    def SetSizer(self, sizer):
        self.sizer = sizer


class MockWxSizer:
    def __init__(self, *args, **kwargs):
        """mock init"""
        self.item = None
        self.margin = None
        self.layout = None
        self.stretch = None

    def Add(self, item, weight, layout, margin):
        self.item = item
        self.margin = margin
        self.layout = layout
        self.stretch = weight


def test_margin_wrapper(mocker):
    """Smoke test.

    If it runs. It is a success.
    """
    mocker.patch.object(wx, "Window", MockWxWindow)
    mocker.patch.object(wx, "BoxSizer", MockWxSizer)

    margins = [(10, wx.LEFT), (10, wx.RIGHT), (1, wx.TOP), (2, wx.BOTTOM)]

    item = MockWxWindow(None)
    window = _margin_wrapper(item, margins)
    print(window)


def test_simple_panel_init(wx_app):
    """Smoke test of initializing the SimplePanel"""
    frame = DefaultFrame("The title")
    panel = SimplePanel(parent=frame)

    assert isinstance(panel.ui_item, wx.Panel)



from __future__ import annotations

import logging
from random import randint

from aio_wx_widgets.controller import BaseController

_LOGGER = logging.getLogger(__name__)


class ControllerThree(BaseController):
    def __init__(self, model):
        self.value_1: int = 0
        self.a_string_value = "A certain string"
        super().__init__(model)

    async def set_value(self):
        val = randint(1, 100)
        _LOGGER.debug("Setting bound property to %s", val)
        self.value_1 = val



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['aio_wx_widgets',
 'aio_wx_widgets.containers',
 'aio_wx_widgets.core',
 'aio_wx_widgets.panels',
 'aio_wx_widgets.widgets']

package_data = \
{'': ['*']}

install_requires = \
['wxasync', 'wxpython', 'pydantic', 'events']

extras_require = \
{'dev': ['pylint', 'nox', 'pytest-cov', 'coverage[toml]>=5.0.3', 'pre-commit'],
 "dev:python_version=='3.7'": ['typing-extensions'],
 'test': ['pytest', 'pytest-asyncio', 'pytest-mock', 'mypy']}

setup(name='aio_wx_widgets',
      version='1.0.4',
      description='aio_wx_widgets',
      author='Sander Teunissen',
      author_email=None,
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



"""Validators"""

import logging
from typing import Union

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "int_validator",
    "float_validator",
    "all_digits_validator",
    "ValidationError",
]


def int_validator(value, force) -> int:  # noqa
    """Int validator."""
    try:
        converted = int(value)
    except ValueError:
        raise ValidationError("Not a valid integer.") from None
    if value == str(converted):
        return converted
    raise ValidationError("not a valid integer.")


def float_validator(value: str, force) -> Union[float, str]:
    """Float validator."""
    try:
        converted = float(value)
    except ValueError:
        raise ValidationError("Not a valid float.") from None
    if value == str(converted):
        return converted
    if force:
        return converted
    return value


def all_digits_validator(value: str, force) -> str:  # noqa
    """All digits validator.

    Only numbers are allowed.
    """
    if value.isdigit():
        return value
    raise ValidationError("Only numbers are allowed.")


class ValidationError(Exception):
    """Input validation error"""



"""A dropdown selection widget."""

from __future__ import annotations

import logging
from typing import Callable, Optional, Sequence, Union

import wx

from aio_wx_widgets import type_annotations as T  # noqa
from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding, TwoWayBindable

_LOGGER = logging.getLogger(__name__)

__all__ = ["Select"]


class Select(BaseWidget):
    """Dropdown selection widget."""

    def __init__(
        self,
        choices: Sequence[T.Choice],
        on_select_callback: Optional[Callable[[T.Choice], None]] = None,
        binding: Optional[T.Binding] = None,
        min_width=300,
        enabled: Union[bool, Binding] = True,
    ):
        """
        Init.

        Args:
            choices: Used to populate the dropdown.
            on_select_callback: called when selection is made.
            binding:
            min_width: when alignment is set, this is the advised minimal width the
                the widget should take.
        """
        value_binding = (
            TwoWayBindable(binding, self._get_ui_value, self._set_ui_value)
            if binding is not None
            else None
        )

        super().__init__(
            wx.Choice(), min_width, value_binding=value_binding, enabled=enabled,
        )
        self.choices = choices
        self._on_select_callback = on_select_callback
        self._selected_item = None

        self.ui_item.Bind(wx.EVT_CHOICE, self._on_choice)
        self.ui_item.Bind(wx.EVT_MOUSEWHEEL, self._on_mouse_wheel)
        self._min_width = min_width

    def _on_mouse_wheel(self, evt):
        """Capturing as I want to disable selection when scrolling"""

    def _set_ui_value(self, value: T.Choice):
        try:
            _idx = next(
                (i for i, val in enumerate(self.choices) if val.value == value.value)
            )
        except StopIteration:
            _LOGGER.error("Default product not available")
        else:
            self.ui_item.Select(_idx)
        # setting this value manually as the SetValue command does -in this case-
        # not trigger the _on_ui_change callback.
        assert self._value_binding is not None
        self._value_binding.fire_update_event = True

    def _on_choice(self, event):  # noqa
        _idx = self.ui_item.GetSelection()
        _LOGGER.debug("Selected idx: %s", _idx)
        self._selected_item = self.choices[_idx]
        if self._value_binding:
            self._value_binding.on_ui_change()
        if self._on_select_callback:
            self._on_select_callback(self._selected_item)

    def _get_ui_value(self, force):  # noqa
        return self._selected_item

    def __call__(self, parent):
        self.ui_item.Create(parent, choices=[choice.label for choice in self.choices])
        self._init()

        return self



"""A panel with a vertical splitter."""

from __future__ import annotations

import logging

import wx
from wx.lib import scrolledpanel

from aio_wx_widgets.core.sizers import SizerMixin

_LOGGER = logging.getLogger(__name__)


class _ScrollPanel(SizerMixin):
    """Panel with scrolling capabilities.

    Used with the TwoSplitterWindow
    """

    def __init__(self, parent):
        self._ui_item = scrolledpanel.ScrolledPanel(parent)
        self._sizer_ = wx.BoxSizer(wx.VERTICAL)
        self._ui_item.SetSizer(self._sizer)
        self._ui_item.SetupScrolling(scroll_x=False)

        super().__init__()

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    @property
    def _sizer(self):
        return self._sizer_

    @property
    def sizer(self):
        """Return the sizer of this panel."""
        return self._sizer


class _SplitterPanel(SizerMixin):
    """Splitter panel.

    Used with the TwoSplitterWindow
    """

    def __init__(self, parent):
        self._ui_item = wx.Panel(parent)
        self._sizer_ = wx.BoxSizer(wx.VERTICAL)
        self.ui_item.SetSizer(self._sizer)

        super().__init__()

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    @property
    def _sizer(self):
        return self._sizer_

    @property
    def sizer(self):
        """Return the sizer of this panel."""
        return self._sizer


class TwoSplitterWindow(SizerMixin):
    """A two splitter window."""

    def __init__(
        self,
        parent,
        window_one_width=250,
        splitter_one_scrollable=True,
        splitter_two_scrollable=True,
    ):
        """Init.

        Args:
            parent:
            window_one_width:
        """

        self._ui_item = wx.SplitterWindow(parent)
        super().__init__()
        if splitter_one_scrollable:
            self.splitter_window_one = _ScrollPanel(self.ui_item)
        else:
            self.splitter_window_one = _SplitterPanel(self.ui_item)
        if splitter_two_scrollable:
            self.splitter_window_two = _ScrollPanel(self.ui_item)
        else:
            self.splitter_window_two = _SplitterPanel(self.ui_item)

        self.ui_item.SetMinimumPaneSize(40)

        self.ui_item.SplitVertically(
            self.splitter_window_one.ui_item,
            self.splitter_window_two.ui_item,
            window_one_width,
        )
        self._sizer_ = self.splitter_window_one.sizer
        self._parent = self.splitter_window_one.ui_item

        self.ui_item.Bind(wx.EVT_SIZE, self._on_resize)

    @property
    def _sizer(self):
        return self._sizer_

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    def _on_resize(self, evt):
        _LOGGER.debug("resizing splitter window.")

        self.splitter_window_one.ui_item.Refresh()
        self.splitter_window_two.sizer.Layout()
        self._sizer.Layout()

        evt.Skip()

    def hide_window2(self):
        """Hide the second window."""
        self.ui_item.Unsplit(self.splitter_window_two.ui_item)

    def show_window2(self):
        """Show the second window."""
        if self.ui_item.IsSplit():
            _LOGGER.debug("Window is already split.")
            return

        self.ui_item.SplitVertically(
            self.splitter_window_one.ui_item, self.splitter_window_two.ui_item
        )

    @property
    def is_split(self):
        """Return whether this window is split or not."""
        return self.ui_item.IsSplit()



import logging
from unittest.mock import Mock

from aio_wx_widgets.core.binding import WATCHERS

_LOGGER = logging.getLogger(__name__)


def test_set_attr_no_watcher(base_controller):
    """Smoke test.

    Check whether setting an attribute "just" works.
    """
    base_controller.a_value = 10

    assert base_controller.a_value == 10


def test_set_attr_watcher_no_subscription(base_controller):
    base_controller.__dict__[WATCHERS] = {}

    base_controller.a_value = 10

    assert base_controller.a_value == 10
    assert "a_value" not in base_controller.__dict__[WATCHERS]


def test_set_attr_watcher_watched(base_controller):
    base_controller.__dict__[WATCHERS] = {}
    mock_watcher = Mock()
    base_controller.__dict__[WATCHERS]["a_value"] = [mock_watcher]

    base_controller.a_value = 10

    mock_watcher.assert_called_with(10)
